import os

from amino.map import Map
from amino import Either


class EnvVars:

    @property
    def vars(self):
        return Map(os.environ)

    def __contains__(self, name: str) -> bool:
        return name in self.vars

    def __getitem__(self, name: str) -> Either[str, str]:
        return self.vars.get(name).to_either(f'env var {name} is unset')

    def __setitem__(self, name: str, value: str) -> None:
        os.environ[name] = str(value)

env = EnvVars()

__all__ = ('EnvVars', 'env')



from typing import TypeVar, Generic, Callable, Union, Any, cast, Optional, Type, Iterator, List as TList, Awaitable
from functools import wraps
from operator import eq
import inspect

from amino import boolean
from amino.tc.base import F
from amino.func import call_by_name, I, curried, CallByName

A = TypeVar('A')
B = TypeVar('B')


class Maybe(Generic[A], F[A], implicits=True):

    __slots__ = ()

    def __new__(cls, value: Optional[A]) -> 'Maybe[A]':
        return Maybe.check(value)

    @staticmethod
    def optional(value: Optional[A]) -> 'Maybe[A]':
        return Nothing if value is None else Just(value)

    @staticmethod
    def check(value: Optional[A]) -> 'Maybe[A]':
        return Maybe.optional(value)

    @staticmethod
    def typed(value: Union[A, B], tpe: Type[A]) -> 'Maybe[A]':
        return Just(value) if isinstance(value, tpe) else Nothing

    @staticmethod
    def wrap(mb: Union['Maybe[A]', None]) -> 'Maybe[A]':
        return mb if mb is not None and isinstance(mb, Just) else Nothing

    @staticmethod
    def getattr(obj: Any, attr: str) -> 'Maybe[A]':
        return Just(getattr(obj, attr)) if hasattr(obj, attr) else Nothing

    @staticmethod
    @curried
    def iff(cond: bool, a: Union[A, Callable[[], A]]) -> 'Maybe[A]':
        return cast(Maybe, Just(call_by_name(a))) if cond else Nothing

    @staticmethod
    @curried
    def iff_m(cond: bool, a: Union['Maybe[A]', Callable[[], 'Maybe[A]']]) -> 'Maybe[A]':
        return cast(Maybe, call_by_name(a)) if cond else Nothing

    @property
    def _get(self) -> Union[A, None]:
        pass

    def cata(self, f: Callable[[A], B], b: Union[B, Callable[[], B]]) -> B:
        return (
            f(cast(A, self._get))
            if self.is_just
            else call_by_name(b)
        )

    def filter(self, f: Callable[[A], bool]) -> 'Maybe[A]':
        l = lambda a: self if f(a) else Nothing
        return self.flat_map(l)

    def get_or_else(self, a: Union[A, Callable[[], A]]) -> A:
        return self.cata(cast(Callable, I), a)

    __or__ = get_or_else

    def get_or_raise(self, e: Exception) -> A:
        def raise_e() -> None:
            raise e
        return self.cata(cast(Callable, I), raise_e)

    def get_or_fail(self, err: CallByName) -> A:
        return self.get_or_raise(Exception(call_by_name(err)))

    def __contains__(self, v: A) -> boolean.Boolean:
        return self.contains(v)

    def error(self, f: Callable[[], Any]) -> 'Maybe[A]':
        self.cata(cast(Callable, I), f)
        return self

    def observe(self, f: Callable[[A], Any]) -> 'Maybe[A]':
        self.foreach(f)
        return self

    effect = observe

    def __iter__(self) -> Iterator:
        return iter(self.to_list)

    @property
    def is_just(self) -> boolean.Boolean:
        return boolean.Boolean(isinstance(self, Just))

    @property
    def is_empty(self) -> boolean.Boolean:
        return ~self.is_just

    empty = is_empty

    @property
    def to_list(self) -> TList[A]:
        from amino.list import List
        return self.cata(lambda v: List(v), List())

    @property
    async def unsafe_await(self) -> 'Maybe[Awaitable]':
        if self.is_just:
            ret = await cast(Callable[[], Awaitable], self._get)()
            return Maybe(ret)
        else:
            return cast(Maybe[Awaitable], self)

    @property
    def contains_coro(self) -> boolean.Boolean:
        return self.exists(inspect.iscoroutine)

    @property
    def json_repr(self) -> Optional[A]:
        return self.cata(cast(Callable, I), lambda: None)


class Just(Generic[A], Maybe[A]):

    __slots__ = 'x',

    def __new__(cls, value: A) -> 'Maybe[A]':
        return object.__new__(cast(Type[object], cls))

    def __init__(self, value: A) -> None:
        self.x = value

    @property
    def _get(self) -> Optional[A]:
        return self.x

    def __str__(self) -> str:
        return 'Just({!s})'.format(self.x)

    def __repr__(self) -> str:
        return 'Just({!r})'.format(self.x)

    def __eq__(self, other: Any) -> bool:
        return eq(self.x, other.x) if isinstance(other, Just) else False

    def __hash__(self) -> int:
        return hash(self._get)


class _Nothing(Generic[A], Maybe[A]):

    __instance: 'Optional[_Nothing]' = None

    def __new__(cls: Type['_Nothing'], *args: Any, **kwargs: Any) -> '_Nothing[A]':
        if _Nothing.__instance is None:
            _Nothing.__instance = object.__new__(cls)
        return cast(_Nothing, _Nothing.__instance)

    def __str__(self) -> str:
        return 'Nothing'

    __repr__ = __str__

    def __eq__(self, other: Any) -> bool:
        return isinstance(other, _Nothing)

    def __hash__(self) -> int:
        return hash('Nothing')

Empty = _Nothing
Nothing: Maybe = _Nothing()


def may(f: Callable[..., Optional[A]]) -> Callable[..., Maybe[A]]:
    @wraps(f)
    def wrapper(*a: Any, **kw: Any) -> Maybe[A]:
        return Maybe.check(f(*a, **kw))
    return wrapper


def flat_may(f: Callable[..., Maybe[A]]) -> Callable[..., Maybe[A]]:
    @wraps(f)
    def wrapper(*a: Any, **kw: Any) -> Maybe[A]:
        res = f(*a, **kw)
        return res if isinstance(res, Maybe) else Nothing
    return wrapper

__all__ = ('Maybe', 'Just', 'Nothing')



import abc
import time
from typing import Union, Any, Callable
from datetime import datetime

import spec

from amino.test.spec import SpecBase, IntegrationSpecBase, default_timeout
from amino.test.sure import SureSpec


def later(ass: Callable[..., bool], *a: Any, timeout: Union[float, None]=None, intval: float=0.1, **kw: Any) -> bool:
    timeout = default_timeout if timeout is None else timeout
    start = datetime.now()
    ok = False
    while not ok and (datetime.now() - start).total_seconds() < timeout:
        try:
            ass(*a, **kw)
            ok = True
        except AssertionError:
            time.sleep(intval)
    return ass(*a, **kw)


class SpecMeta(spec.InnerClassParser, abc.ABCMeta):
    pass


class Spec(SureSpec, SpecBase, spec.Spec, metaclass=SpecMeta):

    def setup(self) -> None:
        SureSpec.setup(self)
        SpecBase.setup(self)

    def _wait_for(self, pred: Callable[[], bool], timeout: float=default_timeout, intval: float=0.1) -> None:
        start = datetime.now()
        while (not pred() and (datetime.now() - start).total_seconds() < timeout):
            time.sleep(intval)
        pred().should.be.ok


class IntegrationSpec(IntegrationSpecBase, Spec):

    def setup(self) -> None:
        IntegrationSpecBase.setup(self)
        Spec.setup(self)

__all__ = ('Spec', 'IntegrationSpec')



import abc
import operator

from toolz import merge

from amino import List, Boolean
from amino.util.fun import format_funcall
from amino.anon.error import AnonError
from amino.list import Lists


class Anon(metaclass=abc.ABCMeta):

    @abc.abstractmethod
    def __call__(self, *a, **kw):
        ...


class AnonCallable(Anon):

    @abc.abstractmethod
    def __call_as_pre__(self, *a, **kw):
        ...

    @abc.abstractproperty
    def __name__(self):
        ...


class AnonAttr(Anon):

    @abc.abstractmethod
    def __substitute_object__(self, obj):
        ...


class AnonMemberCallable(AnonCallable, AnonAttr):
    _call_as_pre_error = 'must pass object to {}'

    def __call_as_pre__(self, *a, **kw):
        def err():
            raise AnonError(self._call_as_pre_error.format(self))
        obj, a2 = List.wrap(a).detach_head.get_or_else(err)
        return self.__call_as_pre_member__(obj, a2)

    def __call_as_pre_member__(self, obj, a):
        return self(obj), a

    def __call__(self, obj):
        return obj


class AnonGetter(AnonMemberCallable):

    def __init__(self, pre, name: str) -> None:
        self.__pre = pre
        self.__name = name

    def __repr__(self):
        return '{}.{}'.format(self.__pre, self.__name)

    def __call_as_pre_member__(self, obj, a):
        return self.__call(obj, a)

    def __call_pre(self, obj, a):
        pre, rest = self.__pre.__call_as_pre__(obj, *a)
        if not hasattr(pre, self.__name):
            raise AttributeError('{!r} has no method \'{}\' -> {!r}'.format(pre, self.__name, self))
        return pre, rest

    def __call(self, obj, a):
        pre, rest = self.__call_pre(obj, a)
        return self.__dispatch__(getattr(pre, self.__name), rest)

    def __call__(self, obj, *a):
        result, rest = self.__call(obj, List.wrap(a))
        if not rest.empty:
            msg = 'extraneous arguments to {}: {}'
            raise AnonError(msg.format(self, rest.mk_string(',')))
        return result

    def __dispatch__(self, obj, a):
        return obj, a

    @property
    def __name__(self):
        return self.__name

    def __substitute_object__(self, obj):
        return self.__pre(obj)


class HasArgs:

    def __substitute__(self, params, args):
        def go(z, arg):
            def transform(value):
                return (arg.__substitute_object__(value)
                        if isinstance(arg, AnonAttr) else value)
            def try_sub(new, rest):
                is_lambda = Boolean(arg is AttrLambdaInst or isinstance(arg, AnonAttr))
                return is_lambda.m(lambda: (transform(new), rest))
            r, a = z
            new, rest = a.detach_head.flat_map2(try_sub) | (arg, a)
            return r.cat(new), rest
        subbed, rest = params.fold_left((List(), args))(go)
        # allow callables to remain in args, but not placeholders
        if rest.empty and subbed.exists(lambda a: a is AttrLambdaInst):
            raise AnonError('too few arguments for {}: {}'.format(self, args))
        return subbed, rest


class AnonFunc(AnonGetter, HasArgs):

    def __init__(self, pre: 'AnonFunc', name: str, a, kw) -> None:
        super().__init__(pre, name)
        self.__args = List.wrap(a)
        self.__kw = kw

    def __dispatch__(self, obj, a):
        sub_a, rest = self.__substitute__(self.__args, List.wrap(a))
        return obj(*sub_a, **self.__kw), rest

    def __getattr__(self, name):
        return MethodRef(self, name)

    def __repr__(self):
        return '{!r}.{}'.format(
            self._AnonGetter__pre,
            format_funcall(self._AnonGetter__name, self.__args, self.__kw)
        )

    def __getitem__(self, key):
        return AnonFunc(self, '__getitem__', [key], {})

    def __substitute_object__(self, obj):
        return self(obj)


class AnonMethod(AnonFunc):
    pass


class MethodRef(AnonMemberCallable):

    def __init__(self, pre: AnonFunc, name: str) -> None:
        self.__pre = pre
        self.__name = name

    def __call__(self, *a, **kw):
        return AnonMethod(self.__pre, self.__name, a, kw)

    def __getattr__(self, name):
        pre = AnonGetter(self.__pre, self.__name)
        return MethodRef(pre, name)

    def __repr__(self):
        return '__.{}'.format(self.__name)

    @property
    def __name__(self):
        return self.__name

    def __substitute_object__(self, obj):
        return getattr(obj, self.__name)

    def __call_as_pre_member__(self, obj, a):
        pre, rest = self.__call_pre(obj, a)
        return self.__dispatch__(getattr(pre, self.__name), rest)


class IdAnonFunc(AnonMemberCallable):

    def __repr__(self):
        return '__'

    @property
    def __name__(self):
        return repr(self)

    def __substitute_object__(self, obj):
        return L(self)


class AnonCall(AnonFunc):

    def __init__(self, pre, a, kw) -> None:
        super().__init__(pre, '__call__', a, kw)

    def __call__(self, obj, *a, **kw):
        return self._AnonGetter__pre(obj)(
            *self._AnonFunc__args, **self._AnonFunc__kw)


class MethodLambda:

    def __getattr__(self, name):
        return MethodRef(IdAnonFunc(), name)

    def __call__(self, *a, **kw):
        return AnonCall(IdAnonFunc(), a, kw)

    def __getitem__(self, key):
        return AnonFunc(IdAnonFunc(), '__getitem__', [key], {})


MethodLambdaInst = MethodLambda()


class AnonFunctionCallable(AnonCallable):
    pass


class AnonChain(AnonFunctionCallable):

    def __init__(self, pre, post) -> None:
        self.__pre = pre
        self.__post = post

    def __call__(self, *a, **kw):
        result, rest = self.__pre.__call_as_pre__(*a, **kw)
        return self.__post(result, *rest, **kw)

    def __call_as_pre__(self, *a, **kw):
        result, rest = self.__pre.__call_as_pre__(*a, **kw)
        return self.__post.__call_as_pre__(*rest, **kw)

    def __str__(self):
        return '({} >> {})'.format(self.__pre, self.__post)


class ComplexLambda(AnonFunctionCallable, HasArgs):

    def __init__(self, func, *a, **kw) -> None:
        assert callable(func), 'ComplexLambda: {} is not callable'.format(func)
        self.__func = func
        self.__args = List.wrap(a)
        self.__kwargs = kw
        self.__qualname__ = self.__func.__name__
        self.__annotations__ = {}

    def __call__(self, *a, **kw):
        result, rest = self.__call(List.wrap(a), kw)
        if not rest.empty:
            msg = 'extraneous arguments to {}: {}'
            raise AnonError(msg.format(self, rest.mk_string(',')))
        return result

    def __call(self, a, kw):
        sub_a, rest = self.__substitute__(self.__args, List.wrap(a))
        sub_kw = merge(self.__kwargs, kw)
        return self.__func(*sub_a, **sub_kw), rest

    def __call_as_pre__(self, *a, **kw):
        return self.__call(a, kw)

    def __getattr__(self, name):
        return MethodRef(self, name)

    @property
    def __name(self):
        return (
            str(self.__func)
            if isinstance(self.__func, AnonCallable) else
            getattr(self.__func, '__name__', str(self.__func))
        )

    def __repr__(self):
        name = 'L({})'.format(self.__name)
        return format_funcall(name, self.__args, self.__kwargs)

    def __rshift__(self, f):
        return AnonChain(self, f)

    @property
    def __name__(self):
        return self.__func.__name__

    def __truediv__(self, other):
        return self.map(other)

    def __floordiv__(self, other):
        return self.flat_map(other)


class LazyMethod(Anon):

    def __init__(self, obj, attr: MethodRef) -> None:
        self.__obj = obj
        self.__attr = attr

    def __call__(self, *a, **kw):
        return self.__attr(*a, **kw)(self.__obj)

    def __getattr__(self, name):
        return LazyMethod(self.__obj, getattr(self.__attr, name))


class ComplexLambdaInst:

    def __init__(self, func) -> None:
        self.__func = func

    def __call__(self, *a, **kw):
        return ComplexLambda(self.__func, *a, **kw)

    def __getattr__(self, name):
        return ComplexLambdaInst(LazyMethod(self.__func, MethodRef(IdAnonFunc(), name)))


def lambda_op(op, s):
    def oper(self, a):
        return OperatorLambda(self.__anon_func__, op, a, s, False)
    return oper


def lambda_rop(op, s):
    def oper(self, a):
        return OperatorLambda(self.__anon_func__, op, a, s, True)
    return oper


class Opers:

    __getitem__ = lambda_op(operator.getitem, "getitem")
    __add__ = lambda_op(operator.add, "+")
    __mul__ = lambda_op(operator.mul, "*")
    __sub__ = lambda_op(operator.sub, "-")
    __mod__ = lambda_op(operator.mod, "%%")
    __pow__ = lambda_op(operator.pow, "**")
    __and__ = lambda_op(operator.and_, "&")
    __or__ = lambda_op(operator.or_, "|")
    __xor__ = lambda_op(operator.xor, "^")
    __div__ = lambda_op(operator.truediv, "/")
    __divmod__ = lambda_op(divmod, "/")
    __floordiv__ = lambda_op(operator.floordiv, "/")
    __truediv__ = lambda_op(operator.truediv, "/")
    __lshift__ = lambda_op(operator.lshift, "<<")
    __rshift__ = lambda_op(operator.rshift, ">>")
    __lt__ = lambda_op(operator.lt, "<")
    __le__ = lambda_op(operator.le, "<=")
    __gt__ = lambda_op(operator.gt, ">")
    __ge__ = lambda_op(operator.ge, ">=")
    __eq__ = lambda_op(operator.eq, "==")
    __ne__ = lambda_op(operator.ne, "!=")
    # __neg__ = unary_lambda_op(operator.neg, "-self")
    # __pos__ = unary_lambda_op(operator.pos, "+self")
    # __invert__ = unary_lambda_op(operator.invert, "~self")
    __radd__ = lambda_rop(operator.add, "+")
    __rmul__ = lambda_rop(operator.mul, "*")
    __rsub__ = lambda_rop(operator.sub, "-")
    __rmod__ = lambda_rop(operator.mod, "%%")
    __rpow__ = lambda_rop(operator.pow, "**")
    __rdiv__ = lambda_rop(operator.truediv, "/")
    __rdivmod__ = lambda_rop(divmod, "/")
    __rtruediv__ = lambda_rop(operator.truediv, "/")
    __rfloordiv__ = lambda_rop(operator.floordiv, "/")
    __rlshift__ = lambda_rop(operator.lshift, "<<")
    __rrshift__ = lambda_rop(operator.rshift, ">>")
    __rand__ = lambda_rop(operator.and_, "&")
    __ror__ = lambda_rop(operator.or_, "|")
    __rxor__ = lambda_rop(operator.xor, "^")


class IdAttrLambda(IdAnonFunc):

    def __repr__(self):
        return '_'


class RootAttrLambda(Opers):

    def __getattr__(self, name):
        return AttrLambda(self.__anon_func__, name)

    @property
    def __anon_func__(self):
        return IdAttrLambda()

    def __repr__(self):
        return '_'


AttrLambdaInst = RootAttrLambda()


class AttrLambda(Opers, AnonGetter, AnonCallable):

    def __init__(self, pre: 'AttrLambda', name: str) -> None:
        super().__init__(pre, name)

    def __getattr__(self, name):
        return AttrLambda(self, name)

    @property
    def __anon_func__(self):
        return self

    def __repr__(self):
        return '{}.{}'.format(self._AnonGetter__pre, self._AnonGetter__name)

    def __substitute_object__(self, obj):
        return self(obj)


class OperatorLambda(AttrLambda):

    def __init__(self, pre: 'AttrLambda', op, strict, name, right) -> None:
        super().__init__(pre, name)
        self.__op = op
        self.__strict = strict
        self.__right = right

    def __call__(self, obj):
        pre = self._AnonGetter__pre(obj)
        a, b = (self.__strict, pre) if self.__right else (pre, self.__strict)
        return self.__op(a, b)

    def __call_as_pre__(self, *a, **kw):
        a0, rest = Lists.wrap(a).detach_head.get_or_fail('no arguments passed to OperatorLambda')
        return self.__call__(a0), rest

    def __repr__(self):
        a, b = (
            (self.__strict, self._AnonGetter__pre)
            if self.__right
            else (self._AnonGetter__pre, self.__strict)
        )
        return '({!r} {} {!r})'.format(a, self._AnonGetter__name, b)

__all__ = ('ComplexLambdaInst', 'RootAttrLambda', 'MethodLambda', 'AttrLambdaInst', 'MethodLambdaInst')



from types import GeneratorType
from typing import TypeVar, Callable, Any, Generator, cast, Optional
import functools

from amino.tc.base import F
from amino.tc.monad import Monad

A = TypeVar('A')
B = TypeVar('B')
G = TypeVar('G', bound=F)


def do(f: Callable[..., Generator[G, B, None]]) -> Callable[..., G]:
    @functools.wraps(f)
    def do_loop(*a: Any, **kw: Any) -> F[B]:
        itr = f(*a, **kw)
        if not isinstance(itr, GeneratorType):
            raise Exception(f'function `{f.__qualname__}` decorated with `do` does not produce a generator')
        c: Optional[F] = None
        m: Optional[Monad[F]] = None
        def send(val: B) -> F[B]:
            nonlocal c, m
            try:
                c = itr.send(val)
                m = Monad.fatal_for(c)
                return c.flat_map(send)
            except StopIteration:
                return m.pure(val)
        return send(cast(B, None))
    return do_loop

__all__ = ('do', 'F')



import abc
import time
import traceback
import inspect
import typing
from typing import Callable, TypeVar, Generic, Any, Union, Tuple, Awaitable

from fn.recur import tco

from amino import Either, Right, Left, Maybe, List, __, Just, env, _, Lists, L, Nothing, options
from amino.eval import Eval
from amino.tc.base import Implicits, ImplicitsMeta
from amino.logging import log
from amino.util.fun import lambda_str, format_funcall
from amino.util.exception import format_exception, sanitize_tb
from amino.util.string import ToStr
from amino.do import do

A = TypeVar('A')
B = TypeVar('B')
C = TypeVar('C')


class IOException(Exception):
    remove_pkgs = List('amino', 'fn')

    def __init__(self, f, stack, cause) -> None:
        self.f = f
        self.stack = List.wrap(stack)
        self.cause = cause

    @property
    def location(self):
        files = List('io', 'anon', 'instances/io', 'tc/base')
        def filt(entry, name):
            return entry.filename.endswith('/amino/{}.py'.format(name))
        stack = self.stack.filter_not(lambda a: files.exists(L(filt)(a, _)))
        pred = (lambda a: not IOException.remove_pkgs
                .exists(lambda b: '/{}/'.format(b) in a.filename))
        return stack.find(pred)

    @property
    def format_stack(self) -> List[str]:
        rev = self.stack.reversed
        def remove_recursion(i):
            pre = rev[:i + 1]
            post = rev[i:].drop_while(__.filename.endswith('/amino/io.py'))
            return pre + post
        def remove_internal():
            start = rev.index_where(_.function == 'unsafe_perform_sync')
            return start / remove_recursion | rev
        frames = (self.location.to_list if IO.stack_only_location else remove_internal())
        data = frames / (lambda a: a[1:-2] + tuple(a[-2]))
        return sanitize_tb(Lists.wrap(traceback.format_list(list(data))))

    @property
    def lines(self) -> List[str]:
        cause = format_exception(self.cause)
        suf1 = '' if self.stack.empty else ' at:'
        tb1 = (List() if self.stack.empty else self.format_stack)
        return tb1.cons(f'IO exception{suf1}').cat('Cause:') + cause + List(
            '',
            'Callback:',
            f'  {self.f}'
        )

    def __str__(self):
        return self.lines.join_lines


class IOMeta(ImplicitsMeta):

    @property
    def zero(self):
        return IO.now(None)


def safe_fmt(f: Callable[..., Any], a: tuple, kw: dict) -> Eval[str]:
    def s() -> str:
        try:
            return format_funcall(f, Lists.wrap(a), kw)
        except Exception as e:
            return str(f)
            log.error(str(e))
    return Eval.later(s)


class IO(Generic[A], Implicits, ToStr, implicits=True, metaclass=IOMeta):
    debug = options.io_debug.exists
    stack_only_location = True

    def __init__(self) -> None:
        self.stack = inspect.stack() if IO.debug else []

    @staticmethod
    def delay(f: Callable[..., A], *a: Any, **kw: Any) -> 'IO[A]':
        return Suspend(L(f)(*a, **kw) >> Pure, safe_fmt(f, a, kw))

    @staticmethod
    def suspend(f: Callable[..., 'IO[A]'], *a: Any, **kw: Any) -> 'IO[A]':
        return Suspend(L(f)(*a, **kw), safe_fmt(f, a, kw))

    @staticmethod
    def call(f: Callable[..., A], *a, **kw):
        return IO.delay(f, *a, **kw)

    @staticmethod
    def now(a: A) -> 'IO[A]':
        return Pure(a)

    @staticmethod
    def pure(a: A) -> 'IO[A]':
        return Pure(a)

    @staticmethod
    def just(a: A) -> 'IO[Maybe[A]]':
        return IO.now(Just(a))

    @staticmethod
    def failed(err: str) -> 'IO[A]':
        def fail():
            raise Exception(err)
        return IO.suspend(fail)

    @staticmethod
    def from_either(a: Either[Any, A]) -> 'IO[A]':
        return a.cata(IO.failed, IO.now)

    @staticmethod
    def from_maybe(a: Maybe[A], error: str) -> 'IO[A]':
        return a / IO.now | IO.failed(error)

    @abc.abstractmethod
    def _flat_map(self, f: Callable[[A], 'IO[B]'], ts: Eval[str], fs: Eval[str]) -> 'IO[B]':
        ...

    @abc.abstractmethod
    def _step(self) -> 'IO[A]':
        ...

    @abc.abstractmethod
    def lambda_str(self) -> Eval[str]:
        ...

    def _arg_desc(self) -> List[str]:
        return List(self.lambda_str()._value())

    def step(self) -> Union[A, 'IO[A]']:
        try:
            return self._step_timed() if IO.debug else self._step()
        except IOException as e:
            raise e
        except Exception as e:
            raise IOException(self.lambda_str()._value(), self.stack, e)

    def run(self) -> A:
        @tco
        def run(t: Union[A, 'IO[A]']) -> Union[Tuple[bool, A], Tuple[bool, Tuple[Union[A, 'IO[A]']]]]:
            if isinstance(t, Pure):
                return True, (t.value,)
            elif isinstance(t, IO):
                return True, (t.step(),)
            else:
                return False, t
        return run(self)

    def flat_map(self, f: Callable[[A], 'IO[B]'], ts: Maybe[Eval[str]]=Nothing, fs: Maybe[Eval[str]]=Nothing
                 ) -> 'IO[B]':
        ts1 = ts | self.lambda_str
        fs1 = fs | Eval.later(lambda: f'flat_map({lambda_str(f)})')
        return self._flat_map(f, ts1, fs1)

    def _step_timed(self) -> Union[A, 'IO[A]']:
        start = time.time()
        v = self._step()
        dur = time.time() - start
        if dur > 0.1:
            log.debug2(lambda: 'IO {} took {:.4f}s'.format(self.string(), dur))
        return v

    def _attempt(self) -> Either[IOException, A]:
        try:
            return Right(self.run())
        except IOException as e:
            return Left(e)

    def unsafe_perform_sync(self) -> Either[IOException, A]:
        return self.attempt

    @property
    def attempt(self) -> Either[IOException, A]:
        return self._attempt()

    @property
    def fatal(self) -> A:
        return self.attempt.get_or_raise

    def and_then(self, nxt: 'IO[B]') -> 'IO[B]':
        fs = Eval.later(lambda: 'and_then({nxt.lambda_str()})')
        return self.flat_map(lambda a: nxt, fs=Just(fs))

    __add__ = and_then

    def join_maybe(self, err: str) -> 'IO[B]':
        def f(a: Union[A, Maybe[B]]) -> 'IO[B]':
            return (
                IO.from_maybe(a, err)
                if isinstance(a, Maybe) else
                IO.failed(f'`IO.join_maybe` called on {self}')
            )
        return self.flat_map(f, fs=Just(Eval.later('join_maybe({})'.format, err)))

    @property
    def join_either(self) -> 'IO[B]':
        def f(a: Union[A, Either[C, B]]) -> 'IO[B]':
            return (
                IO.from_either(a)
                if isinstance(a, Either) else
                IO.failed(f'`IO.join_either` called on {self}')
            )
        return self.flat_map(f, fs=Just(Eval.now('join_either')))

    def with_stack(self, s: typing.List[inspect.FrameInfo]) -> 'IO[A]':
        self.stack = s
        return self

    def recover(self, f: Callable[[IOException], B]) -> 'IO[B]':
        return IO.delay(lambda: self.attempt).map(__.value_or(f))

    @do
    def ensure(self, f: Callable[[Either[IOException, A]], 'IO[None]']) -> 'IO[A]':
        result = yield IO.delay(lambda: self.attempt)
        yield f(result)
        yield IO.from_either(result)

    @property
    def coro(self) -> Awaitable[Either[IOException, A]]:
        async def coro() -> Either[IOException, A]:
            return self.attempt
        return coro()


class Suspend(Generic[A], IO[A]):

    def __init__(self, thunk: Callable[[], IO[A]], string: Eval[str]) -> None:
        super().__init__()
        self.thunk = thunk
        self.string = string

    def lambda_str(self) -> Eval[str]:
        return self.string

    def _step(self) -> IO[A]:
        return self.thunk().with_stack(self.stack)

    def _flat_map(self, f: Callable[[A], IO[B]], ts: Eval[str], fs: Eval[str]) -> IO[B]:
        return BindSuspend(self.thunk, f, ts, fs)


class BindSuspend(Generic[A], IO[A]):

    def __init__(self, thunk: Callable[[], IO[A]], f: Callable, ts: Eval[str], fs: Eval[str]) -> None:
        super().__init__()
        self.thunk = thunk
        self.f = f
        self.ts = ts
        self.fs = fs

    def lambda_str(self) -> Eval[str]:
        return (self.ts & self.fs).map2('{}.{}'.format)

    def _step(self) -> IO[A]:
        return (
            self.thunk()
            .flat_map(self.f, fs=Just(self.fs))
            .with_stack(self.stack)
        )

    def _flat_map(self, f: Callable[[A], IO[B]], ts: Eval[str], fs: Eval[str]) -> IO[B]:
        bs = L(BindSuspend)(self.thunk, lambda a: self.f(a).flat_map(f, Just(ts), Just(fs)), ts, fs)
        return Suspend(bs, (ts & fs).map2('{}.{}'.format))


class Pure(Generic[A], IO[A]):

    def __init__(self, value: A) -> None:
        super().__init__()
        self.value = value

    def lambda_str(self) -> Eval[str]:
        return Eval.later(lambda: f'Pure({self.value})')

    def _arg_desc(self) -> List[str]:
        return List(str(self.value))

    def _step(self) -> IO[A]:
        return self

    def _flat_map(self, f: Callable[[A], IO[B]], ts: Eval[str], fs: Eval[str]) -> IO[B]:
        return Suspend(L(f)(self.value), (ts & fs).map2('{}.{}'.format))


def Try(f: Callable[..., A], *a: Any, **kw: Any) -> Either[Exception, A]:
    try:
        return Right(f(*a, **kw))
    except Exception as e:
        return Left(e)


def io(fun: Callable[..., A]) -> Callable[..., IO[A]]:
    def dec(*a: Any, **kw: Any) -> IO[A]:
        return IO.delay(fun, *a, **kw)
    return dec

__all__ = ('IO', 'Try', 'io')



from typing import TypeVar, Callable, Tuple, Optional, cast
from functools import reduce

from amino import maybe, List, Maybe
from amino.func import curried
from amino.lazy import lazy
from amino.tc.monad import Monad
from amino.tc.base import ImplicitInstances, tc_prop
from amino.tc.traverse import Traverse, TraverseF, TraverseG
from amino.tc.applicative import Applicative
from amino.tc.foldable import Foldable, FoldableABC
from amino.list import flatten
from amino.tc.zip import Zip
from amino.tc.monoid import Monoid

A = TypeVar('A', covariant=True)
B = TypeVar('B')


class ListInstances(ImplicitInstances):

    @lazy
    def _instances(self):
        from amino import Map
        return Map(
            {
                Monad: ListMonad(),
                Traverse: ListTraverse(),
                Foldable: ListFoldable(),
                Zip: ListZip(),
                Monoid: ListMonoid(),
            }
        )


class ListMonad(Monad):

    def pure(self, b: B) -> List[B]:
        return List(b)

    def map(self, fa: List[A], f: Callable[[A], B]) -> List[B]:
        return List.wrap(map(f, fa))

    def flat_map(self, fa: List[A], f: Callable[[A], List[B]]) -> List[B]:
        return List.wrap(flatten(map(f, fa)))


class ListTraverse(Traverse):

    def traverse(self, fa: List[A], f: Callable, tpe: type):
        monad = Applicative.fatal(tpe)
        def folder(z, a):
            return monad.map2(z.product(f(a)), lambda l, b: l.cat(b))
        return fa.fold_left(monad.pure(List()))(folder)


FoldableABC.register(List)
TraverseF.register(List)
TraverseG.register(List)


def _find(fa: List[A], f: Callable[[A], bool]) -> Optional[A]:
    return next(filter(f, fa), None)


class ListFoldable(Foldable):

    @tc_prop
    def with_index(self, fa: List[A]) -> List[Tuple[int, A]]:
        return List.wrap(enumerate(fa))

    def filter(self, fa: List[A], f: Callable[[A], bool]):
        return List.wrap(filter(f, fa))

    @curried
    def fold_left(self, fa: List[A], z: B, f: Callable[[B, A], B]) -> B:
        return reduce(f, fa, z)

    def find(self, fa: List[A], f: Callable[[A], bool]):
        return Maybe(_find(fa, f))

    def find_map(self, fa: List[A], f: Callable[[A], Maybe[B]]) -> Maybe[B]:
        for el in fa:
            found = f(el)
            if found.present:
                return found
        return maybe.Empty()

    def index_where(self, fa: List[A], f: Callable[[A], bool]):
        gen = (maybe.Just(i) for i, a in enumerate(fa) if f(a))
        return next(gen, maybe.Nothing)


class ListZip(Zip):

    def zip(self, fa: List[A], fb: List[B], *fs) -> List:
        return List.wrap(zip(fa, fb, *fs))


class ListMonoid(Monoid):

    @property
    def empty(self):
        return List()

    def combine(self, fa: List, fb: List):
        return fa + fb

__all__ = ('ListInstances',)



from amino.anon.prod.method import MethodLambdaInst
from amino.anon.prod.attr import AttrLambdaInst
from amino.anon.prod.complex import ComplexLambdaInst

__all__ = ('MethodLambdaInst', 'AttrLambdaInst', 'ComplexLambdaInst')



from typing import Any, cast

from amino import options
from amino.util.mod import unsafe_import_name
from amino.anon.prod.attr import AttrLambda
from amino.anon.prod.method import MethodLambda
from amino.anon.prod.complex import ComplexLambda

_: AttrLambda = cast(AttrLambda, None)
__: MethodLambda = cast(MethodLambda, None)
L: ComplexLambda = cast(ComplexLambda, None)


def set(mod: str) -> None:
    def name(name: str) -> Any:
        return unsafe_import_name(mod, name)
    global _, __, L
    _ = name('AttrLambdaInst')
    __ = name('MethodLambdaInst')
    L = name('ComplexLambdaInst')


def set_debug() -> None:
    set('amino.anon.debug')


def set_prod() -> None:
    set('amino.anon.prod')

if options.anon_debug:
    set_debug()
else:
    set_prod()

__all__ = ('set_debug', 'set_prod', '_', '__', 'L')



import re
import abc
from typing import Any, Callable
from functools import singledispatch

from hues import huestr

import amino


def snake_case(name: str) -> str:
    s1 = re.sub('(.)([A-Z][a-z]+)', r'\1_\2', name)
    return re.sub('([a-z0-9])([A-Z])', r'\1_\2', s1).lower()


@singledispatch
def decode(value: Any) -> Any:
    return value


@decode.register(bytes)
def decode_bytes(value: bytes) -> str:
    return value.decode()


@decode.register(list)
def decode_list(value: list) -> 'amino.List[str]':
    return amino.List.wrap(value).map(decode)


@decode.register(dict)
def decode_dict(value: dict) -> 'amino.Map[str, str]':
    return amino.Map.wrap(value).keymap(decode).valmap(decode)


@decode.register(Exception)
def decode_exc(value: Exception) -> str:
    return decode_list(value.args).head | str(value)


def camelcase(name: str, sep: str='', splitter: str='_') -> str:
    return sep.join([n.capitalize() for n in re.split(splitter, name)])

camelcaseify = camelcase


def safe_string(value: Any) -> str:
    try:
        return str(value)
    except Exception:
        try:
            return repr(value)
        except Exception:
            return 'invalid'


class ToStr(abc.ABC):

    @abc.abstractmethod
    def _arg_desc(self) -> 'amino.List[str]':
        ...

    def __str__(self) -> str:
        args = self._arg_desc().join_comma
        return f'{self.__class__.__name__}({args})'

    def __repr__(self) -> str:
        return str(self)


def col(a: Any, c: Callable[[huestr], huestr]) -> str:
    return c(huestr(str(a))).colorized


def red(a: Any) -> str:
    return col(a, lambda a: a.red)


def green(a: Any) -> str:
    return col(a, lambda a: a.green)


def yellow(a: Any) -> str:
    return col(a, lambda a: a.yellow)


def blue(a: Any) -> str:
    return col(a, lambda a: a.blue)


def cyan(a: Any) -> str:
    return col(a, lambda a: a.cyan)


def magenta(a: Any) -> str:
    return col(a, lambda a: a.magenta)

__all__ = ('snake_case', 'decode', 'camelcaseify', 'red', 'green', 'yellow', 'blue', 'cyan', 'magenta', 'camelcase')



from amino import List


async def gather_sync(data, f):
    results = List()
    for el in data:
        results = results + List(await f(el))
    return results


async def gather_sync_flat(data, f):
    results = List()
    for el in data:
        results = results + (await f(el)).to_list
    return results

__all__ = ('gather_sync', 'gather_sync_flat')



import abc
import importlib
from typing import TypeVar, Generic, Callable, Union, Any, cast, Iterator, Type
from types import ModuleType  # noqa
from pathlib import Path

import amino  # noqa
from amino import boolean
from amino.func import I
from amino.tc.base import F
from amino.util.mod import unsafe_import_name
from amino.tc.monoid import Monoid
from amino.util.string import ToStr
from amino.do import do
from amino.util.exception import format_exception

A = TypeVar('A')
B = TypeVar('B')
C = TypeVar('C')
E = TypeVar('E', bound=Exception)


class ImportFailure(ToStr):

    @abc.abstractproperty
    def expand(self) -> 'amino.List[str]':
        ...


class ImportException(ImportFailure):

    def __init__(self, desc: str, exc: Exception) -> None:
        self.desc = desc
        self.exc = exc

    def _arg_desc(self) -> 'amino.List[str]':
        return amino.List(self.desc, str(self.exc))

    @property
    def expand(self) -> 'amino.List[str]':
        return format_exception(self.exc).cons(self.desc)


class InvalidLocator(ImportFailure):

    def __init__(self, msg: str) -> None:
        self.msg = msg

    def _arg_desc(self) -> 'amino.List[str]':
        return amino.List(self.msg)

    @property
    def expand(self) -> 'amino.List[str]':
        return amino.List(self.msg)


class Either(Generic[A, B], F[B], implicits=True):

    def __init__(self, value: Union[A, B]) -> None:
        self.value = value

    @staticmethod
    def import_name(mod: str, name: str) -> 'Either[ImportFailure, B]':
        try:
            value = unsafe_import_name(mod, name)
        except Exception as e:
            return Left(ImportException(f'{mod}.{name}', e))
        else:
            return Left(InvalidLocator(f'{mod} has no attribute {name}')) if value is None else Right(value)

    @staticmethod
    def import_path(path: str) -> 'Either[ImportFailure, B]':
        from amino.list import List
        return (
            List.wrap(path.rsplit('.', 1))
            .lift_all(0, 1)
            .to_either(InvalidLocator(f'invalid module path: {path}'))
            .flat_map2(lambda a, b: Either.import_name(a, b).lmap(lambda a: ImportException(path, a)))
        )

    @staticmethod
    def import_module(modname: str) -> 'Either[ImportFailure, ModuleType]':
        try:
            mod = importlib.import_module(modname)
        except Exception as e:
            return Left(ImportException(modname, e))
        else:
            return Right(mod)

    @staticmethod
    def import_file(path: Path) -> 'Either[ImportFailure, ModuleType]':
        from amino.maybe import Maybe
        def step2(spec: importlib._bootstrap.ModuleSpec) -> ModuleType:
            module = importlib.util.module_from_spec(spec)
            spec.loader.exec_module(module)
            return module
        try:
            module = Maybe.check(importlib.util.spec_from_file_location('temp', str(path))) / step2
        except Exception as e:
            return Left(ImportException(str(path), e))
        else:
            return module.to_either(InvalidLocator(f'failed to import `{path}`'))

    @staticmethod
    @do
    def import_from_file(path: Path, name: str) -> 'Either[ImportFailure, B]':
        module = yield Either.import_file(path)
        attr = getattr(module, name, None)
        yield (
            Left(InvalidLocator(f'{path} has no attribute {name}'))
            if attr is None else
            Right(attr)
        )

    @staticmethod
    def catch(f: Callable[[], B], exc: Type[E]) -> 'Either[E, B]':
        try:
            return Right(f())
        except exc as e:
            return Left(e)

    @property
    def is_right(self) -> 'amino.Boolean':
        return boolean.Boolean(isinstance(self, Right))

    @property
    def is_left(self) -> 'amino.Boolean':
        return boolean.Boolean(isinstance(self, Left))

    @property
    def __left_value(self) -> A:
        return cast(A, self.value)

    @property
    def __right_value(self) -> B:
        return cast(B, self.value)

    def leffect(self, f: Callable[[A], Any]) -> 'Either[A, B]':
        if self.is_left:
            f(self.__left_value)
        return self

    def bieffect(self, l: Callable[[A], Any],
                 r: Callable[[B], Any]) -> 'Either[A, B]':
        self.cata(l, r)
        return self

    def cata(self, fl: Callable[[A], C], fr: Callable[[B], C]) -> C:
        return fl(self.__left_value) if self.is_left else fr(self.__right_value)

    def bimap(self, fl: Callable[[A], 'Either[C, D]'], fr: Callable[[B], 'Either[C, D]']) -> 'Either[C, D]':
        return Left(fl(self.__left_value)) if self.is_left else Right(fr(self.__right_value))

    def recover_with(self, f: Callable[[A], 'Either[C, B]']) -> 'Either[C, B]':
        return self.cata(f, Right)

    def right_or_map(self, f: Callable[[A], C]) -> C:
        return self.cata(f, I)

    def value_or(self, f: Callable[[A], B]) -> B:
        return self.cata(f, I)

    def left_or(self, f: Callable[[B], A]) -> A:
        return self.cata(I, f)

    def left_or_map(self, f: Callable[[B], A]) -> A:
        return self.cata(I, f)

    @property
    def ljoin(self) -> 'Either[A, C]':
        return self.right_or_map(Left)

    def __str__(self) -> str:
        return '{}({})'.format(self.__class__.__name__, self.value)

    def __repr__(self) -> str:
        return '{}({!r})'.format(self.__class__.__name__, self.value)

    @property
    def to_list(self) -> 'amino.List[B]':
        return self.to_maybe.to_list

    def lmap(self, f: Callable[[A], C]) -> 'Either[C, B]':
        return cast(Either, Left(f(self.__left_value))) if self.is_left else cast(Either, Right(self.__right_value))

    @property
    def get_or_raise(self) -> B:
        def fail(err: A) -> B:
            raise err if isinstance(err, Exception) else Exception(err)
        return self.cata(fail, I)

    @property
    def fatal(self) -> B:
        return self.get_or_raise

    def __iter__(self) -> Iterator[B]:
        return iter(self.to_list)

    @property
    def swap(self) -> 'Either[B, A]':
        return self.cata(Right, Left)

    @property
    def json_repr(self) -> B:
        return self.to_maybe.json_repr

    def accum_error(self, b: 'Either[A, C]') -> 'Either[A, C]':
        return self.accum_error_f(lambda: b)

    def accum_error_f(self, f: Callable[[], 'Either[A, C]']) -> 'Either[A, C]':
        def acc(v: A) -> None:
            return Monoid.fatal(type(v)).combine(self.__left_value, v)
        return f().lmap(acc) if self.is_left else self

    def filter_with(self, f: Callable[[B], bool], g: Callable[[B], C]) -> 'Either[C, B]':
        return self // (lambda a: Right(a) if f(a) else Left(g(a)))

    def left_contains(self, a: A) -> 'boolean.Boolean':
        return boolean.Boolean(self.is_left and self.__left_value == a)


class Right(Either):

    def __eq__(self, other: Any) -> bool:
        return isinstance(other, Right) and self._Either__right_value == other._Either__right_value


class Left(Either):

    def __eq__(self, other: Any) -> bool:
        return isinstance(other, Left) and self._Either__left_value == other._Either__left_value

__all__ = ('Either', 'Left', 'Right', 'ImportFailure', 'ImportException', 'InvalidLocator')



import typing
from typing import GenericMeta, Any
from types import SimpleNamespace

from amino import List, Lists
from amino.util.string import ToStr


class AlgebraMeta(GenericMeta):

    def __new__(cls, name: str, bases: typing.List[type], namespace: SimpleNamespace, base: bool=False, **kw: Any
                ) -> None:
        @property
        def sub(self) -> List[AlgebraMeta]:
            if self._sub is None:
                def intermediate(tpe: type) -> bool:
                    return tpe.__name__ == self.__name__
                return Lists.wrap(self.__subclasses__()).filter_not(intermediate)
            return self._sub
        if base:
            cls._sub = None
            cls.sub = sub
        return super().__new__(cls, name, bases, namespace, **kw)


class Algebra(ToStr, metaclass=AlgebraMeta):
    pass

__all__ = ('AlgebraMeta', 'Algebra')



import abc
from typing import Generic, TypeVar, Callable, Tuple, Any, Union

from fn.recur import tco

from amino.tc.base import F
from amino.tc.monad import Monad
from amino import List

A = TypeVar('A')
B = TypeVar('B')


class Eval(Generic[A], F[A], implicits=True, auto=True):

    @staticmethod
    def now(a: A) -> 'Eval[A]':
        return Now(a)

    @staticmethod
    def later(f: Callable[..., A], *a: Any, **kw: Any) -> 'Eval[A]':
        return Later(lambda: f(*a))

    @staticmethod
    def always(f: Callable[..., A], *a: Any, **kw: Any) -> 'Eval[A]':
        return Always(lambda: f(*a))

    @abc.abstractmethod
    def _value(self) -> A:
        ...

    @property
    def value(self) -> A:
        return self._value()

    @abc.abstractproperty
    def memoize(self) -> 'Eval[A]':
        ...


class Now(Generic[A], Eval[A]):

    def __init__(self, value: A) -> None:
        self.strict = value

    def _value(self) -> A:
        return self.strict

    @property
    def memoize(self) -> Eval[A]:
        return self

    def __str__(self) -> str:
        return f'Now({self.strict})'


class Later(Generic[A], Eval[A]):

    def __init__(self, f: Callable[[], A]) -> None:
        self.f = f
        self._memo = None  # type: Union[A, None]

    def _memoized(self) -> A:
        if self._memo is None:
            self._memo = self.f()
        return self._memo

    def _value(self) -> A:
        return self._memoized()

    @property
    def memoize(self) -> Eval[A]:
        return self

    def __str__(self) -> str:
        return f'Later({self.f})'


class Always(Generic[A], Eval[A]):

    def __init__(self, f: Callable[[], A]) -> None:
        self.f = f

    def _value(self) -> A:
        return self.f()

    @property
    def memoize(self) -> Eval[A]:
        return Later(self.f)

    def __str__(self) -> str:
        return f'Always({self.f})'


class Call(Generic[A], Eval[A]):

    def __init__(self, thunk: Callable[[], Eval[A]]) -> None:
        self.thunk = thunk

    @staticmethod
    def _loop(ev: Eval[A]) -> Eval[A]:
        def loop1(s: Eval[A]) -> Eval[A]:
            return loop(s.run)
        @tco
        def loop(e: Eval[A]) -> Tuple[bool, Tuple[Eval[A]]]:
            if isinstance(e, Call):
                return True, (e.thunk(),)
            elif isinstance(e, Compute):
                return False, (Compute(e.start, loop1, e.start_str),)
            else:
                return False, e
        return loop(ev)

    @property
    def memoize(self) -> Eval[A]:
        return Later(lambda: self._value())

    def _value(self) -> A:
        return Call._loop(self).value

    def __str__(self) -> str:
        return f'Call({self.thunk})'


class Compute(Generic[A, B], Eval[A]):

    def __init__(self, start: Callable[[], Eval[B]], run: Callable[[B], Eval[A]], start_str: str) -> None:
        self.start = start
        self.run = run
        self.start_str = start_str

    def _value(self) -> A:
        C = Callable[[Any], Eval[Any]]
        def loop_compute(c: Compute[Any, Any], fs: List[C]) -> Tuple[bool, Union[Tuple[Eval[Any], List[C]], Any]]:
            cc = c.start()
            return (
                (True, (cc.start(), fs.cons(c.run).cons(cc.run)))
                if isinstance(cc, Compute) else
                (True, (c.run(cc._value()), fs))
            )
        def loop_other(e: Eval[Any], fs: List[C]) -> Tuple[bool, Union[Tuple[Eval[Any], List[C]], Any]]:
            return fs.detach_head.map2(lambda fh, ft: (True, (fh(e._value()), ft))) | (False, e._value())
        @tco
        def loop(e: Eval[Any], fs: List[C]) -> Tuple[bool, Union[Tuple[Eval[Any], List[C]], Any]]:
            return (
                loop_compute(e, fs)
                if isinstance(e, Compute) else
                loop_other(e, fs)
            )
        return loop(self, List())

    @property
    def memoize(self) -> Eval[A]:
        return Later(self)

    def __str__(self) -> str:
        return f'Compute({self.start_str})'


class EvalMonad(Monad, tpe=Eval):

    def pure(self, a: A) -> Eval[A]:
        return Now(a)

    def flat_map(self, fa: Eval[A], f: Callable[[A], Eval[B]]) -> Eval[B]:
        def f1(s: A) -> Eval[B]:
            return Compute(lambda: fa.run(s), f, '')
        start, run, start_str = (
            (fa.start, f1, fa.start_str)
            if isinstance(fa, Compute) else
            (fa.thunk, f, fa.thunk_str)
            if isinstance(fa, Call) else
            (lambda: fa, f, str(fa))
        )
        return Compute(start, run, start_str)


def defer_eval(f: Callable[[], Eval[A]]) -> Eval[A]:
    return Call(f)

__all__ = ('Eval', 'defer_eval')



import abc
from typing import TypeVar, Generic

from amino.tc.base import TypeClass

F = TypeVar('F')


class Monoidal(Generic[F], TypeClass[F]):

    @abc.abstractmethod
    def product(self, fa: F, fb: F) -> F:
        ...

__all__ = ('Monoidal',)



import sys
from typing import Union

from amino import env, Path

__base_dir__ = None


class TestEnvError(Exception):
    pass


def setup(path: Union[str, Path]) -> None:
    ''' Use the supplied path to initialise the tests base dir.
    If _path is a file, its dirname is used.
    '''
    if not isinstance(path, Path):
        path = Path(path)
    if not path.is_dir():
        path = path.parent
    global __base_dir__
    __base_dir__ = path
    container = str(pkg_dir())
    if container not in sys.path:
        sys.path.insert(0, container)
        env['PYTHONPATH'] = '{}:{}'.format(container, env['PYTHONPATH'] | '')


def _check() -> Path:
    if __base_dir__ is None:
        msg = 'Test base dir not set! Call amino.test.setup(dir).'
        raise TestEnvError(msg)
    else:
        return __base_dir__


def temp_path(*components: str) -> Path:
    dir = _check()
    return Path(dir, '_temp', *components)


def temp_dir(*components: str) -> Path:
    _dir = temp_path(*components)
    _dir.mkdir(exist_ok=True, parents=True)
    return _dir


def temp_file(*components: str) -> Path:
    return temp_dir(*components[:-1]).joinpath(*components[-1:])


def create_temp_file(*components: str) -> Path:
    _file = temp_file(*components)
    _file.touch()
    return _file


def fixture_path(*components: str) -> Path:
    dir = _check()
    return Path(dir, '_fixtures', *components)


def load_fixture(*components: str) -> str:
    with fixture_path(*components).open() as f:
        return f.read()


def base_dir() -> Path:
    return _check()


def pkg_dir() -> Path:
    return base_dir().parent

__all__ = ('create_temp_file', 'temp_file', 'temp_path', 'temp_dir', 'fixture_path', 'load_fixture', 'base_dir',
           'pkg_dir')
